%%%----------------------------------------------------------------------
%%% File    : cron_server.erl
%%% Author  : ottuser local account <otpuser@tiger>
%%% Purpose : Perform scheduled backups of mnesia
%%% Created :  1 Oct 1999 by ottuser local account <otpuser@tiger>
%%%----------------------------------------------------------------------

-module(cron_server).
-author('otpuser@tiger').
-vsn('$Id: cron_server.erl,v 1.1.1.1 2001/04/30 09:46:24 chandru Exp $ ').

-behaviour(gen_server).

%% External exports
-export([start_link/1]).

%% gen_server callbacks
-export([init/1, handle_call/3, handle_cast/2, handle_info/2, terminate/2, code_change/3]).

-export([rescan/0, do_apply/3, print/0]).



%% Crontab Format
%% {{Daysofmonth, Daysofweek, Hours, mins}, Startups mfa, Regular mfa, Finish mfa}.
%% {{'*', '*', '*', '*'}, {log_mgr, open, [csa_actions, "csa_actions.log", "/export/home/optuser/app_logs"]},
%%		          {log_mgr, wrap, [csa_actions, "csa_actions.log", "/export/home/optuser/app_logs", 5]},
%%                        {log_mgr, close, [csa_actions]}}.

%% State is list of {Times, Regular, End, Value}
%% e.g. [{{'*','*', [1,2], '*'}, {io, format, ["Regular~n"]}, {io, format, ["End~n"]}, '$false'}]

%%%----------------------------------------------------------------------
%%% API
%%%----------------------------------------------------------------------
start_link(Crontab) ->
    gen_server:start_link({local, cron_server}, ?MODULE, [Crontab], []).

rescan() ->
    gen_server:call(cron_server, rescan).

print() ->
    gen_server:call(cron_server, print).
%%%----------------------------------------------------------------------
%%% Callback functions from gen_server
%%%----------------------------------------------------------------------

%%----------------------------------------------------------------------
%% Func: init/1
%% Returns: {ok, State}          |
%%          {ok, State, Timeout} |
%%          ignore               |
%%          {stop, Reason}
%%----------------------------------------------------------------------
init(Crontab) ->
    process_flag(trap_exit, true),
    {ok, State} = do_startups(Crontab),
    {ok, State, timeout()}.

%%----------------------------------------------------------------------
%% Func: handle_call/3
%% Returns: {reply, Reply, State}          |
%%          {reply, Reply, State, Timeout} |
%%          {noreply, State}               |
%%          {noreply, State, Timeout}      |
%%          {stop, Reason, Reply, State}   | (terminate/2 is called)
%%          {stop, Reason, State}            (terminate/2 is called)
%%----------------------------------------------------------------------
handle_call(rescan, From, State) ->
    Crontab = case application:get_env(crontab) of
		  {ok, Tab} ->
		      Tab;
		  undefined ->
		      filename:join(code:priv_dir(cron), "cron.conf")
	      end,
    ok = do_finish(State),
    {ok, New_state} = do_startups(Crontab),
    {reply, ok, New_state, timeout()};
handle_call(print, From, State) ->
    {reply, State, State, timeout()};
handle_call(Call, From, State) ->
    {reply, ok, State, timeout()}.


%%----------------------------------------------------------------------
%% Func: handle_cast/2
%% Returns: {noreply, State}          |
%%          {noreply, State, Timeout} |
%%          {stop, Reason, State}            (terminate/2 is called)
%%----------------------------------------------------------------------
handle_cast(Msg, State) ->
    {noreply, State, timeout()}.

%%----------------------------------------------------------------------
%% Func: handle_info/2
%% Returns: {noreply, State}          |
%%          {noreply, State, Timeout} |
%%          {stop, Reason, State}            (terminate/2 is called)
%%----------------------------------------------------------------------
handle_info(timeout, State) ->
    do_actions(calendar:local_time(), State),
    {noreply, State, timeout()};

handle_info(Info, State) ->
    {noreply, State, timeout()}.

%%----------------------------------------------------------------------
%% Func: terminate/2
%% Purpose: Shutdown the server
%% Returns: any (ignored by gen_server)
%%----------------------------------------------------------------------
terminate(Reason, State) ->
    do_finish(State).

%%----------------------------------------------------------------------
%% Func: code_change/3
%% Purpose: Convert process state when code is changed
%% Returns: {ok, NewState}
%%----------------------------------------------------------------------
code_change(OldVsn, State, Extra) ->
    {ok, State, timeout()}.

%%%----------------------------------------------------------------------
%%% Internal functions
%%%----------------------------------------------------------------------

%%%----------------------------------------------------------------------
%%% Startup Routine
%%%----------------------------------------------------------------------
do_startups(Crontab) ->
    {ok, Entries} = file:consult(Crontab),
    {ok, State} = start_each(Entries, []).

start_each([{Times, {Ms, Fs, As}, Regular, End}|Rest], Acc) ->
    case catch apply(Ms, Fs, As) of
	ok ->
	    start_each(Rest, [{Times, Regular, End, '$false'}|Acc]);
	{ok, Value} ->
	    start_each(Rest, [{Times, Regular, End, Value}|Acc]);
	{'EXIT', Reason} ->
	    error("Startup routine failed.~nEntry:~p~nReason:~p", [{Ms, Fs, As},Reason]),
	    start_each(Rest, Acc);
	Else ->
	    error("Startup Routine returned invalid result.~nEntry:~p~nReason:~p", [{Ms, Fs, As},Else]),
	    start_each(Rest, Acc)
    end;
start_each([{Times, false, Regular, End}|Rest], Acc) ->
    start_each(Rest, [{Times, Regular, End, '$false'}|Acc]);
start_each([], Acc) ->
    {ok, Acc}.

%%%----------------------------------------------------------------------
%%% Action Routine. Run through all of the entries comparing Now with the
%%% spec. Do jobs which match
%%%----------------------------------------------------------------------
do_actions({{Y,M,D}, {Ho,Mi,Se}}, State) ->
    Dow=calendar:day_of_the_week(Y,M,D),
    action_each({Dow,Y,M,D,Ho,Mi}, State).

action_each({Dow,Y,Mo,D,Ho,Mi}, [{{Mo_spec, We_spec, Ho_spec, Mi_spec}, {M, F, A}, _, Value}|T]) ->
    case if_action(match(Mi, Mi_spec, {0, 59}),
		   match(Ho, Ho_spec, {0, 23}),
		   match(Dow, We_spec, {1, 7}),
		   match(D, Mo_spec, {1, calendar:last_day_of_the_month(Y, Mo)})) of
	true ->
	    do_apply(M, F, val(Value) ++ A),
	    action_each({Dow,Y,Mo,D,Ho,Mi}, T);
	false ->
	    action_each({Dow,Y,Mo,D,Ho,Mi}, T)
    end;
action_each(_, []) ->
    ok.

%% match when '*' and in range, or when in list AND in range
match(Val, '*', {Start, End}) when Val >= Start, Val =< End -> true;
match(Val, List, {Start, End}) when Val >= Start, Val =< End -> lists:member(Val, List);
match(Val, List, {Start, End}) -> false.

%% Hours and minutes must match together with one or both of Day of week and Day of month
if_action(true, true, true, true) -> true;
if_action(true, true, true, false) -> true;
if_action(true, true, false, true) -> true;
if_action(_, _, _, _) -> false.

val('$false') -> [];
val(Val) when list(Val) -> Val;
val(Val) -> [Val].

do_apply(M, F, A) ->
    case catch apply(M, F, A) of
	{ok, Result} ->
	    ok;
	ok ->
	    ok;
	{'EXIT', Reason} ->
	    error("Cron Job Failed: ~p~nReason:~p~n", [{M, F, A}, Reason]);
	Else ->
	    error("Cron Job Failed: ~p~nReturned:~p~n", [{M, F, A}, Else])
    end.

%%%----------------------------------------------------------------------
%%% Finish Routine
%%%----------------------------------------------------------------------
do_finish(State) ->
    lists:foreach(fun({_, _, {M, F, A}, Value}) ->
			  Args = val(Value) ++ A,
			  catch apply(M, F, Args),
			  ok
		  end, State).

%%%----------------------------------------------------------------------
%%% Work out how long it is until the next :00 minute
%%%----------------------------------------------------------------------
timeout() ->
    {_, Time} = calendar:local_time(), 
    Time_since_last =  calendar:time_to_seconds(Time) rem 60,
    (60-Time_since_last)*1000.


error(Text, Content) ->
    Text2 = lists:flatten(io_lib:format(Text, Content)),
    error_logger:error_report([{application, cron}, Text2]).
